from __future__ import annotations

from collections import Counter
from typing import Iterable, List, Optional, Union

from .utils import make_protein, make_sequences, indexcheck
from collections import abc

class Sequence(abc.Sequence):

    def __init__(self, 
                 sequence: str,
                 rna: bool = False) -> None:
        self.sequence: str = sequence
        self.rna: bool = rna
        self._prot: str = ''

    @indexcheck
    def __getitem__(self, index: int) -> Union[Sequence, str]:
        if isinstance(index, int):
            return self.sequence[index]
        else:   # if index is tuple of indices
            return Sequence(sequence=''.join(self[i] for i in range(*index)),
                            rna=self.rna)

    def __contains__(self, x: object) -> bool:
        if not isinstance(x,  (Sequence, str)):
            raise TypeError(f'Must be str or Sequence')
        if isinstance(x, str):
            return x in self.sequence
        return x.sequence in self.sequence

    def __repr__(self) -> str:
        return self.sequence

    def __str__(self) -> str:
        return self.sequence

    def __len__(self) -> int:
        return len(self.sequence)

    def index(self, other: object) -> int:
        if not isinstance(other, (str, Sequence)):
            raise ValueError
        if isinstance(other, Sequence):
            return self.sequence.index(other.sequence)
        return self.sequence.index(other)


    @property
    def nucleotide_counter(self) -> Counter:
        return Counter(self.sequence)

    
    def make_frames(self, 
                    start_codons: Iterable[str],
                    stop_codons: Iterable[str],
                    len_seq: int = 50) -> Optional[List[Sequence]]:
        seqs: List[str] = make_sequences(seq=self.sequence,
                                         start_codons=start_codons,
                                         stop_codons=stop_codons,
                                         len_seq=len_seq)
        if seqs:
            return [Sequence(seq, rna=self.rna) for seq in seqs]
        return []
    

    @property
    def protein(self) -> str:
        if not self._prot:
            if len(self.sequence) % 3:
                raise ValueError(f'Length of sequence must be divided by 3')
            self._prot = make_protein(self.sequence, rna=self.rna)
        return self._prot


    def transcription(self) -> str:
        if self.rna:
            return self.sequence.replace('U', 'T')
        return self.sequence.replace('T', 'U')


    def aminoacids(self) -> Counter:
        return Counter(self.protein)
        


from collections import defaultdict, Counter
from itertools import product, chain
from typing import Dict, List

from ..sequence import Sequence

from .utils import get_ambiguous_nucleotides, make_nucleotides, validation


class Metasequence:

    def __init__(self, sequence: str, rna: bool = False) -> None:
        validation(seq=sequence)
        self.seq: str = sequence        
        self.rna = rna
        self._amb_nucleotides: Dict[str, List[int]] = []
        self._content: Dict[int, str] = defaultdict(str)
        n = 0
        level: List[int] = [n]
        self._levels: List[List[int]] = []
        next_level: List[int] = []

        for nucl in sequence:
            nucleotides: List[str] = make_nucleotides(nucl)
            if len(nucleotides) == 1 and len(level) == 1:
                self._content[n] += nucleotides[0]
            else:
                for nucl_ in nucleotides:
                    n += 1
                    next_level.append(n)
                    self._content[n] += nucl_
                self._levels.append(level)
                level = next_level
                next_level = []
        if level not in chain.from_iterable(self._levels):
            self._levels.append(level)

    def make_sequences(self) -> List[Sequence]:
        sequences: List[Sequence] = []
        paths = product(*self._levels)
        for path in paths:
            seq = ''
            for node in path:
                seq += self._content[node]
            sequences.append(Sequence(seq, rna=self.rna))
        return sequences

    @property
    def ambiguous_nucleotides(self):
        if not self._amb_nucleotides:
            self._amb_nucleotides = get_ambiguous_nucleotides(self.seq)
        return self._amb_nucleotides

    @property
    def count_nucleotides(self) -> Counter:
        Counter(self.seq)



from typing import Dict, List
from collections import defaultdict
from ..codon_dict import iupac_nucleotides

def make_nucleotides(nucl: str) -> List[str]:
    return [i for i in iupac_nucleotides[nucl]]

def validation(seq: str) -> None:
    diff = set(seq) - set(iupac_nucleotides.keys())
    if diff:
        raise ValueError(f'There are invalid nucleotides in sequence: {diff}. Or invalid path {seq}')

def get_ambiguous_nucleotides(seq) -> Dict[str, List[int]]:
    ambiguous_nucleotides: Dict[str, List[int]] = defaultdict(list)
    for i, nucl in enumerate(seq):
        if not nucl in 'ACGTU':
            ambiguous_nucleotides[nucl].append(i)
    return dict(ambiguous_nucleotides)



from typing import Iterable, List, Set
from ..codon_dict import rna_codons, dna_codons


def make_sequences(seq: str, start_codons: Iterable[str], stop_codons: Iterable[str], len_seq: int) -> List[str]:
    start: Set[str] = set(start_codons)
    stop: Set[str] = set(stop_codons)
    seqs = []
    for i in range(len(seq)):
        if seq[i: i+3] in start:
            for j in range(i+3, len(seq), 3):
                if seq[j: j+3] in stop:
                    if len(seq[i:j]) > len_seq:
                        seqs.append(seq[i:j])
                    break
    return seqs


def make_protein(sequence, rna) -> str:
        if rna:
            codons = rna_codons
        else:
            codons = dna_codons
        protein: str = ''
        for i in range(0, len(sequence), 3):
            triplet: str = sequence[i:i+3]
            protein += codons[triplet]
        return protein

def indexcheck(f):
    def inner(self, index, *args):
        if isinstance(index, int):
            if index < 0:
                index += len(self)
            if not (0 <= index < len(self)):
                raise IndexError('list index out of range')
        elif isinstance(index, slice):
            index = index.indices(len(self))
        else:
            raise TypeError('Not valid type of index')

        return f(self, index, *args)
    return inner



'''pycodon'''

__version__ = '0.1.1'
import os

from .metasequence import Metasequence
from .utils import data_from_file

def read_file(filepath: str):
    if not os.path.isfile(filepath):
        raise ValueError(f'Path to file {filepath} is invalid')
    data: str = data_from_file(filepath)
    return Metasequence(data)



import pathlib
from setuptools import find_packages, setup


HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
    name='pycodon',
    version='0.1.1',
    python_requires='>=3.8.0',
    description='read DNA sequences',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/tikeeva/pycodon',
    author=['ETikeeva', 'DMalakhova', 'DZanadvornykh'],
    author_email='tikeeva_es@mail.ru',
    license='MIT',
    classifiers=['License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.8',

    ],
    packages=find_packages(exclude=('scripts',)),
    include_package_data=True, 
    entry_points={
        "console_scripts": []}
    
)


def data_from_file(file: str) -> str:
    with open(file) as f:
        first_row: str = f.readline()
        if first_row.startswith('>'):
            first_row = f.readline()
        next_row: str = f.readline()
        while next_row:
            first_row += next_row
            next_row = f.readline()
    return first_row.replace('\n', '').replace(' ', '')
